import copy
import ast
import collections.abc
import datalink.links
import logging
from datalink.utils import GenericEntry, ListEntry
from traits.api import HasTraits


log = logging.getLogger(__name__)


class DataStore(HasTraits):
    """Class for a basic mapping data store."""
    url = None
    table = None
    _fields = {}
    _lookup = None
    _bidirectional = True

    def __init__(self, *args, **kwargs):

        if args and not len(args) == 1:
            raise ValueError('Only takes K0 or 1 positional arguments.')

        # Intercept user defined values
        instance_map = copy.deepcopy(self._fields)
        datastore_arg_dict = {}
        for k, v in kwargs.items():
            if k in instance_map:
                datastore_arg_dict[k] = v
        for k in datastore_arg_dict:
            kwargs.pop(k)
        instance_map.update(datastore_arg_dict)

        # Set the traits
        for attr, value in instance_map.items():
            trait = f'{attr}_trait'
            if isinstance(value, collections.abc.Iterable) and not isinstance(value, str):
                setattr(self, trait, ListEntry(value))
            else:
                setattr(self, trait, GenericEntry(value))
            getattr(self, trait).on_trait_change(
                self._conditional_save_state, 'val[]'
            )

        # Flags for internal operation.
        self._save_flag = True

        # Establish link
        link_id = None
        if args:
            link_id = args[0]
        self.link = self._get_link(link_id)

        # Check for any found data and initialise it.
        if self.link.loaded_data:
            self._format_loaded_data()
            # If already exists and the user passed
            # args, update the entry.
            if datastore_arg_dict:
                self.update(**datastore_arg_dict)

        # Save new entries.
        else:
            self._conditional_save_state()

    def __getattr__(self, name):
        """Intercept traits."""
        if name in self._fields:
            if self._bidirectional:
                self._force_load()
            return getattr(self, f'{name}_trait').val
        else:
            raise AttributeError(name)

    def __setattr__(self, attr, value):
        if attr in self._fields:
            self.__dict__[f'{attr}_trait'].val = value
        else:
            self.__dict__[attr] = value

    def __bool__(self):
        """Return true if data for the id was found in the database."""
        if self.link.loaded_data:
            return True
        return False

    # Properties for interfacing with the link to save, and to handle
    # translation between SQL friendly data and the python objects in
    # the data store.
    def _get_link(self, link_id):
        """Factory method to construct the link."""
        if self._lookup == 'uuid':  # Only uuid supported at present.
            return datalink.links.UUIDLookup(table_name=self.table,
                                             url=self.url,
                                             link_id=link_id)
        else:
            raise ValueError(self._lookup)

    def _force_load(self):
        self.link.loaded_data = self.link.load()
        self._format_loaded_data()

    def _conditional_save_state(self):
        """Push an update to the db only if the _save_flag is true."""
        if self._save_flag:
            self.link.save(self._sql_friendly_data)

    def save(self):
        """Method to force a save."""
        self.link.save(self._sql_friendly_data)

    @property
    def data(self):
        """Property exposing the most up-to-date version of the data."""
        self._force_load()
        return self._data

    @property
    def _data(self):
        d = {k: v for k, v in zip(self._fields,
                                  [getattr(self, f'{attr}_trait').val
                                   for attr in self._fields])}
        d['id'] = self.id
        return d

    @property
    def _sql_friendly_data(self):
        """
        Property to return a version of the data store
        with data types supported by SQL backends.
        """
        d = copy.deepcopy(self._data)
        for key, val in d.items():
            if isinstance(val, collections.abc.Sequence) and not isinstance(val, str):
                try:
                    d[key] = str(val)
                except TypeError:
                    raise
        # Add the uuid
        d['id'] = self.link.id
        return d

    def _format_loaded_data(self):
        """
        Take the loaded data and format it back into python objects.
        Should only be called in initialisation.
        """
        d = self.link.loaded_data
        d.pop('id')
        self._save_flag = False
        for k, v in d.items():
            try:
                v = ast.literal_eval(v)
            except (ValueError, SyntaxError):
                pass
            setattr(self, k, v)
        self._save_flag = True

    @property
    def id(self):
        return self.link.id

    def update(self, **kwargs):
        """
        Update multiple properties at once with one push to db.
        """
        self._save_flag = False
        for k in kwargs:
            if k not in self._data:
                raise KeyError(f'update received a non data store parameter: {k}')
        for i, (k, v) in enumerate(kwargs.items()):
            if i == len(kwargs) - 1:
                self._save_flag = True
            setattr(self, k, v)



from traits.api import Int, Float, Str, List, HasTraits, Tuple, Generic, Dict


class GenericEntry(HasTraits):
    val = Generic()
    def __init__(self, v):
        self.val = v


# Classes to make traits work.
class ListEntry(HasTraits):
    val = List()
    def __init__(self, v):
        self.val = v


class DictEntry(HasTraits):
    val = Dict()
    def __init__(self, v):
        self.val = v



# -*- coding: utf-8 -*-

__author__ = 'Stephen Ogilvy'
__copyright__ = 'Stephen Ogilvy'
__licence__ = 'GPLv3'
__version__ = '0.0.1'
__email__ = 'sogilvy@protonmail.com'

import logging
import sqlalchemy
import types
import collections.abc
import datalink.links as dllinks
import datalink.stores as dlstores
import datalink.utils as dlutils


log = logging.getLogger(__name__)


def test_output():
    log.info('Test logging output from datalink.')


def factory(
        name, table, fields,
        url=None,
        database=None,
        lookup='uuid', bidirectional=True,
        ):
    """
    Factory function to produce a new class derived from DataStore.
    """
    # Check args and kwargs
    for arg in [name, fields, table]:
        if not arg:
            raise ValueError(f'{arg} is a required positional field.')
    if not isinstance(fields, collections.abc.Mapping):
        raise ValueError('fields must be a valid mapping.')

    # If url not supplied, assume sqlite and use database as file path
    # to construct an sqlite URL using the default sqlalchemy driver.
    if not url:
        url = sqlalchemy.engine.url.URL('sqlite', database=database)

    new_class = types.new_class(name, bases=(dlstores.DataStore, ))
    new_class.__name__ = name

    new_class.url = str(url)
    new_class.table = table
    new_class._fields = fields

    new_class._lookup = lookup
    new_class._bidirectional = bidirectional
    return new_class



import abc
import dataset
import logging
import uuid
import sqlalchemy
import sqlalchemy_utils


log = logging.getLogger(__name__)


class SQLInterface:
    """Class to handle all interactions with SQL databases."""

    def __init__(self, table_name='data', url=None,
                 link_id=None, **kwargs):
        self._table = table_name
        self._id = link_id
        self._url = url
        self.loaded_data = None

        self.ensure_database()
        self.ensure_table()

        try:
            self._lookup_dict = kwargs['lookup']
        except KeyError:
            pass

        # Try to load.
        if link_id:
            self.loaded_data = self.load()
            if self.loaded_data:
                log.debug(f'Loaded data corresponding to ID: {self.id}')

    def ensure_database(self):
        """Ensure the database for the type of data exists."""
        if not sqlalchemy_utils.database_exists(self.url):
            try:
                s = sqlalchemy_utils.create_database(self.url)
                log.info(f'db created at: {self.url}')
            except sqlalchemy.exc.SQLAlchemyError as e:
                log.error(f'failed to create db at: {self.url}')
                log.error(e)
                raise

    def ensure_table(self):
        with dataset.connect(self.url) as db:
            if self.table not in db.tables:
                t = db[self.table]
                log.info(f'created table {self.table}')

    @property
    @abc.abstractmethod
    def id(self):
        pass

    @property
    def url(self):
        return self._url

    @property
    def table(self):
        return self._table

    def load(self):
        """Method to attempt a load from the relevant table."""
        with dataset.connect(self.url) as db:
            t = db[self.table]
            results = list(t.find(id=str(self.id)))
            if not results:
                return None
            if len(results) > 1:
                log.warning(f'Ambiguous uuid in loading of data,'
                            f' received {len(results)} results. Taking first')
            return results[0]

    def save(self, data):
        """Method to save or update the relevant entry in the table."""
        if not self.load():
            with dataset.connect(self.url) as db:
                log.debug(f'Creating new database entry with id {self.id}.')
                if self.table in db.tables:
                    t = db[self.table]
                else:
                    t = db.create_table(self.table, primary_id=False)
                t.insert(data)
        else:
            with dataset.connect(self.url) as db:
                log.debug(f'Updating existing database entry for id {self.id}.')
                t = db[self.table]
                t.update(data, ['id'])


class UUIDLookup(SQLInterface):
    """A lookup interface with a uuid as auto-generated entry identifier."""

    @property
    def id(self):
        if not self._id:
            self._id = uuid.uuid4()
        return str(self._id)



from setuptools import setup, find_packages
from os import path
here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='datalink',
    version='0.1.0',

    description=('Create simple interfaces to SQL that make working with data as simple'
                ' as working with dictionaries.'),

    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hypernormalisation/datalink',
    author='Stephen Ogilvy',
    author_email='sogilvy@protonmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Database :: Front-Ends',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    keywords='development database sql',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    # python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    python_requires='>=3.3',
    install_requires=['sqlalchemy',
                      'dataset',
                      'sqlalchemy',
                      'sqlalchemy_utils',
                      'traits'],

    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    # project_urls={  # Optional
    #     'Bug Reports': 'https://github.com/pypa/sampleproject/issues',
    #     'Funding': 'https://donate.pypi.org',
    #     'Say Thanks!': 'http://saythanks.io/to/example',
    #     'Source': 'https://github.com/pypa/sampleproject/',
    # },
)
